package habibviewer.jpeg2000;

import jj2000.j2k.quantization.dequantizer.*;
import jj2000.j2k.image.invcomptransf.*;
import jj2000.j2k.fileformat.reader.*;
import jj2000.j2k.codestream.reader.*;
import jj2000.j2k.wavelet.synthesis.*;
import jj2000.j2k.entropy.decoder.*;
import jj2000.j2k.codestream.*;
import jj2000.j2k.image.*;
import jj2000.j2k.util.*;
import jj2000.j2k.roi.*;
import jj2000.j2k.io.*;
import jj2000.disp.*;
import jj2000.j2k.*;
import jj2000.j2k.decoder.DecoderSpecs;

import colorspace.*;

import java.awt.image.*;
import java.util.*;
import java.awt.*;
import java.net.*;
import java.io.*;

import icc.ICCProfiler;

/**
 * This class is alteration of the main class of JJ2000's decoder. It instantiates all
 * objects and performs the decoding operations. It then writes the image to
 * the output file or displays it.
 *
 * <p>First the decoder should be initialized with a input file name and loading parameters
 * given through the constructor. When the run() method is invoked the
 * decoder executes. The exit code of the class can be obtained with the
 * getExitCode() method, after the constructor and after the run method. A
 * non-zero value indicates that an error has ocurred.</p>
 *
 * <p>The decoding chain corresponds to the following sequence of modules:</p>
 *
 * <ul>
 * <li>BitstreamReaderAgent</li>
 * <li>EntropyDecoder</li>
 * <li>ROIDeScaler</li>
 * <li>Dequantizer</li>
 * <li>InverseWT</li>
 * <li>ImgDataConverter</li>
 * <li>EnumratedColorSpaceMapper, SyccColorSpaceMapper or ICCProfiler</li>
 * <li>ComponentDemixer (if needed)</li>
 * <li>ImgDataAdapter (if ComponentDemixer is needed)</li>
 * <li>BlkImgDataSrcImageProducer</li>
 * </ul>
 *
 * <p>The behaviour of each module may be modified according to the current
 * tile-component. All the specifications are kept in modules extending
 * ModuleSpec and accessible through an instance of DecoderSpecs class.</p>
 *
 * @see BitstreamReaderAgent
 * @see EntropyDecoder
 * @see ROIDeScaler
 * @see Dequantizer
 * @see InverseWT
 * @see ImgDataConverter
 * @see InvCompTransf
 * @see BlkImgDataSrcImageProducer
 * @see ModuleSpec
 * @see DecoderSpecs
 * */
public class Decoder implements Runnable {

    /** Parses the inputstream to analyze the box structure of the JP2
     * file. */
    private ColorSpace csMap = null;

    /** Reference to the TitleUpdater instance. Only used when decoded image
     * is displayed */
    TitleUpdater title = null;

    /** The exit code of the run method */
    private int exitCode;

    /** The parameter list (arguments) */
    private ParameterList pl;

    /** Information contained in the codestream's headers */
    private HeaderInfo hi;

    /** Frame used to display decoded image */
    private Frame win = null;

    /** The component where the image is to be displayed */
    private ImgScrollPane isp;

    /** Input file*/
    private String infile;

    /** Error output stream */
    public String errorMessage = "";

    /** Decoded image */
    public BlkImgDataSrc decodedImage;

    /** Number of layers */
    public int nLayers = -1;

    /** Number of Components in codestream */
    public int nCompCod = -1;

    /** Number of tiles */
    public int nTiles = -1;

    /** Target image width */
    public int targetImageWidth = -1;

    /** Target image height */
    public int targetImageHeight = -1;

    /** Tile width */
    public int tileWidth = -1;

    /** Tile height */
    public int tileHeight = -1;

    /** Reconstructed resolution level */
    public int reconstructedResolutionLevel = -1;

    /** Target resolution level */
    public int targetResolutionLevel = -1;

    /** Reconstructed image width */
    public int reconstructedImageWidth = -1;

    /** Reconstructed image height */
    public int reconstructedImageHeight = -1;

    /** Target bitrate */
    public float targetBitrate = -1;

    /** The parameter information for this class */
    private final static String[][] pinfo = {
        {"u", "[on|off]",
         "Prints usage information. " +
            "If specified all other arguments (except 'v') are ignored", "off"},
        {"v", "[on|off]",
         "Prints version and copyright information", "off"},
        {"verbose", "[on|off]",
         "Prints information about the decoded codestream", "on"},
        {"pfile", "<filename>",
         "Loads the arguments from the specified file. Arguments that are " +
            "specified on the command line override the ones from the file.\n" +
            "The arguments file is a simple text file with one argument per " +
            "line of the following form:\n" +
            "  <argument name>=<argument value>\n" +
            "If the argument is of boolean type (i.e. its presence turns a " +
            "feature on), then the 'on' value turns it on, while the 'off' " +
            "value turns it off. The argument name does not include the '-' " +
            "or '+' character. Long lines can be broken into several lines " +
            "by terminating them with '\\'. Lines starting with '#' are " +
            "considered as comments. This option is not recursive: any 'pfile' " +
            "argument appearing in the file is ignored.", null},
        {"res", "<resolution level index>",
         "The resolution level at which to reconstruct the image " +
            " (0 means the lowest available resolution whereas the maximum " +
            "resolution level corresponds to the original image resolution). " +
            "If the given index" +
            " is greater than the number of available resolution levels of the " +
            "compressed image, the image is reconstructed at its highest " +
            "resolution (among all tile-components). Note that this option" +
            " affects only the inverse wavelet transform and not the number " +
            " of bytes read by the codestream parser: this number of bytes " +
            "depends only on options '-nbytes' or '-rate'.", null},
        {"i", "<filename or url>",
         "The file containing the JPEG 2000 compressed data. This can be " +
            "either a JPEG 2000 codestream or a JP2 file containing a " +
            "JPEG 2000 " +
            "codestream. In the latter case the first codestream in the file " +
            "will be decoded. If an URL is specified (e.g., http://...) " +
            "the data will be downloaded and cached in memory before decoding. " +
            "This is intended for easy use in applets, but it is not a very " +
            "efficient way of decoding network served data.", null},
        {"o", "<filename>",
         "This is the name of the file to which the decompressed image " +
            "is written. If no output filename is given, the image is " +
            "displayed on the screen. " +
            "Output file format is PGX by default. If the extension" +
            " is '.habibviewer.pgm' then a PGM file is written as output, however this is " +
            "only permitted if the component bitdepth does not exceed 8. If " +
            "the extension is '.habibviewer.ppm' then a PPM file is written, however this " +
            "is only permitted if there are 3 components and none of them has " +
            "a bitdepth of more than 8. If there is more than 1 component, " +
            "suffices '-1', '-2', '-3', ... are added to the file name, just " +
            "before the extension, except for PPM files where all three " +
            "components are written to the same file.", null},
        {"rate", "<decoding rate in bpp>",
         "Specifies the decoding rate in bits per pixel (bpp) where the " +
            "number of pixels is related to the image's original size (Note:" +
            " this number is not affected by the '-res' option). If it is equal" +
            "to -1, the whole codestream is decoded. " +
            "The codestream is either parsed (default) or truncated depending " +
            "the command line option '-parsing'. To specify the decoding " +
            "rate in bytes, use '-nbytes' options instead.", "-1"},
        {"nbytes", "<decoding rate in bytes>",
         "Specifies the decoding rate in bytes. " +
            "The codestream is either parsed (default) or truncated depending " +
            "the command line option '-parsing'. To specify the decoding " +
            "rate in bits per pixel, use '-rate' options instead.", "-1"},
        {"parsing", null,
         "Enable or not the parsing mode when decoding rate is specified " +
            "('-nbytes' or '-rate' options). If it is false, the codestream " +
            "is decoded as if it were truncated to the given rate. If it is " +
            "true, the decoder creates, truncates and decodes a virtual layer" +
            " progressive codestream with the same truncation points in each " +
            "code-block.", "on"},
        {"ncb_quit", "<max number of code blocks>",
         "Use the ncb and lbody quit conditions. If state information is " +
            "found for more code blocks than is indicated with this option, " +
            "the decoder " +
            "will decode using only information found before that point. " +
            "Using this otion implies that the 'rate' or 'nbyte' parameter " +
            "is used to indicate the lbody parameter which is the number of " +
            "packet body bytes the decoder will decode.", "-1"},
        {"l_quit", "<max number of layers>",
         "Specifies the maximum number of layers to decode for any code-" +
            "block", "-1"},
        {"m_quit", "<max number of bit planes>",
         "Specifies the maximum number of bit planes to decode for any code" +
            "-block", "-1"},
        {"poc_quit", null,
         "Specifies the whether the decoder should only decode code-blocks " +
            "included in the first progression order.", "off"},
        {"one_tp", null,
         "Specifies whether the decoder should only decode the first " +
            "tile part of each tile.", "off"},
        {"comp_transf", null,
         "Specifies whether the component transform indicated in the " +
            "codestream should be used.", "on"},
        {"debug", null,
         "Print debugging messages when an error is encountered.", "off"},
        {"cdstr_info", null,
         "Display information about the codestream. This information is: " +
            "\n- Marker segments value in main and tile-part headers," +
            "\n- Tile-part length and position within the code-stream.", "off"},
        {"nocolorspace", null,
         "Ignore any colorspace information in the image.", "off"},
        {"colorspace_debug", null,
         "Print debugging messages when an error is encountered in the" +
            " colorspace module.", "off"}
    };

    /**
     * Instantiates a decoder object, with the input file name string given as
     * argument, a component where to display the image and decoding parameters,
     * like target bitrate, number of layers, number of bit planes and number
     * of code blocks.
     *
     * @param isp The component where the image is to be displayed
     *
     * @param inputFile Name of the input file
     *
     * @param rate Decoding rate in bits per pixel
     *
     * @param nbytes Decoding rate in bytes
     *
     * @param layers Maximum number of layers to decode for any code-block
     *
     * @param bitplanes Maximum number of bit planes to decode for any code-block
     *
     * @param codeblocks Maximum number of code-blocks
     * */
    public Decoder(ImgScrollPane isp, String inputFile, float rate, int nbytes,
                   int layers, int bitplanes, int codeblocks) {
        this.isp = isp;
        this.infile = inputFile;

        ParameterList defaultParameterList;
        defaultParameterList = new ParameterList();
        String[][] param = getAllParameters();
        for (int i = param.length - 1; i >= 0; i--) {
            if (param[i][3] != null)
                defaultParameterList.put(param[i][0], param[i][3]);
        }
        pl = new ParameterList(defaultParameterList);
        pl.put("i", inputFile);
        pl.put("parsing", "off");

        if (rate != -1) pl.put("rate", String.valueOf(rate)); else
        if (nbytes != -1) pl.put("nbytes", String.valueOf(nbytes));

        if (layers > 0) pl.put("l_quit", Integer.toString(layers + 1));
        pl.put("ncb_quit", Integer.toString(codeblocks));
        pl.put("m_quit", Integer.toString(bitplanes));

    }

    /**
     * Returns the exit code of the class. This is only initialized after the
     * constructor and when the run method returns.
     *
     * @return The exit code of the constructor and the run() method.
     * */
    public int getExitCode() {
        return exitCode;
    }

    /**
     * Returns the parameters that are used in this class. It returns a 2D
     * String array. Each of the 1D arrays is for a different option, and they
     * have 3 elements. The first element is the option name, the second one
     * is the synopsis and the third one is a long description of what the
     * parameter is. The synopsis or description may be 'null', in which case
     * it is assumed that there is no synopsis or description of the option,
     * respectively.
     *
     * @return the options name, their synopsis and their explanation.
     * */
    private static String[][] getParameterInfo() {
        return pinfo;
    }

    /**
     * Runs the decoder. After completion the exit code is set, a non-zero
     * value indicates that an error ocurred.
     *
     * @see #getExitCode
     * */
    public void run() {

        int res; // resolution level to reconstruct
        RandomAccessIO in;
        FileFormatReader ff;
        BitstreamReaderAgent breader;
        HeaderDecoder hd;
        EntropyDecoder entdec;
        ROIDeScaler roids;
        Dequantizer deq;
        InverseWT invWT;
        InvCompTransf ictransf;
        ImgDataConverter converter;
        DecoderSpecs decSpec = null;
        BlkImgDataSrc palettized;
        BlkImgDataSrc channels;
        BlkImgDataSrc resampled;
        BlkImgDataSrc color;
        int i;
        int depth[];
        Image img = null;

        try {
            if (infile == null) {
                error("Input file has not been specified", 1);
                return;
            }

            // **** Open input files ****
            // Creates a BEBufferedRandomAccessFile of a ISRandomAccessIO
            // instance for reading the file format and codestream data
            if (infile.indexOf("/") >= 1 &&
                    infile.charAt(infile.indexOf("/") - 1) == ':') { // an URL
                URL inurl;
                URLConnection conn;
                int datalen;
                InputStream is;

                try {
                    inurl = new URL(infile);
                } catch (MalformedURLException e) {
                    error("Malformed URL for input file " + infile, 4);
                    return;
                }
                try {
                    conn = inurl.openConnection();
                    conn.connect();
                } catch (IOException e) {
                    error("Cannot open connection to " + infile +
                            ((e.getMessage() != null) ?
                            (":\n" + e.getMessage()) : ""), 4);
                    return;
                }
                datalen = conn.getContentLength();
                try {
                    is = conn.getInputStream();
                } catch (IOException e) {
                    error("Cannot get data from connection to " + infile +
                            ((e.getMessage() != null) ?
                            (":\n" + e.getMessage()) : ""), 4);
                    return;
                }
                if (datalen != -1) { // known length => initialize to length
                    in = new ISRandomAccessIO(is, datalen, 1, datalen);
                } else { // unknown length => use defaults
                    in = new ISRandomAccessIO(is);
                }
                // HACK: to verify if the URL is valid try to read some data
                try {
                    in.read();
                    in.seek(0);
                } catch (IOException e) {
                    error("Cannot get input data from " + infile +
                            " Invalid URL?", 4);
                    return;
                }
            } else { // a normal file
                try {
                    in = new BEBufferedRandomAccessFile(infile, "r");
                } catch (IOException e) {
                    error("Cannot open input file " +
                            ((e.getMessage() != null) ?
                            (":\n" + e.getMessage()) : ""), 4);
                    return;
                }
            }

            // **** File Format ****
            // If the codestream is wrapped in the jp2 fileformat, Read the
            // file format wrapper
            ff = new FileFormatReader(in);
            ff.readFileFormat();
            if (ff.JP2FFUsed) {
                in.seek(ff.getFirstCodeStreamPos());
            }

            // +----------------------------+
            // | Instantiate decoding chain |
            // +----------------------------+

            // **** Header decoder ****
            // Instantiate header decoder and read main header
            hi = new HeaderInfo();
            try {
                hd = new HeaderDecoder(in, pl, hi);
            } catch (EOFException e) {
                error("Codestream too short or bad header, " +
                        "unable to decode.", 2);
                return;
            }

            decSpec = hd.getDecoderSpecs();

            // Report information
            nLayers = decSpec.nls.getMax();
            nCompCod = hd.getNumComps();
            nTiles = hi.siz.getNumTiles();
            targetImageWidth = hi.siz.getCompImgWidth(0);
            targetImageHeight = hi.siz.getCompImgHeight(0);
            tileWidth = hi.siz.xtsiz;
            tileHeight = hi.siz.ytsiz;

            // Get demixed bitdepths
            depth = new int[nCompCod];
            for (i = 0; i < nCompCod; i++) {
                depth[i] = hd.getOriginalBitDepth(i);
            }

            // **** Bit stream reader ****
            try {
                breader = BitstreamReaderAgent.
                        createInstance(in, hd, pl, decSpec,
                                pl.getBooleanParameter("cdstr_info"), hi);
            } catch (IOException e) {
                error("Error while reading bit stream header or parsing " +
                        "packets" + ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 4);
                return;
            } catch (IllegalArgumentException e) {
                error("Cannot instantiate bit stream reader" +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }

            // **** Entropy decoder ****
            try {
                entdec = hd.createEntropyDecoder(breader, pl);
            } catch (IllegalArgumentException e) {
                error("Cannot instantiate entropy decoder" +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }

            // **** ROI de-scaler ****
            try {
                roids = hd.createROIDeScaler(entdec, pl, decSpec);
            } catch (IllegalArgumentException e) {
                error("Cannot instantiate roi de-scaler." +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }

            // **** Dequantizer ****
            try {
                deq = hd.createDequantizer(roids, depth, decSpec);
            } catch (IllegalArgumentException e) {
                error("Cannot instantiate dequantizer" +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }

            // **** Inverse wavelet transform ***
            try {
                // full page inverse wavelet transform
                invWT = InverseWT.createInstance(deq, decSpec);
            } catch (IllegalArgumentException e) {
                error("Cannot instantiate inverse wavelet transform" +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }

            res = breader.getImgRes();
            invWT.setImgResLevel(res);

            // **** Data converter **** (after inverse transform module)
            converter = new ImgDataConverter(invWT, 0);

            // **** Inverse component transformation ****
            ictransf = new InvCompTransf(converter, decSpec, depth, pl);

            // **** Color space mapping ****
            if (ff.JP2FFUsed && pl.getParameter("nocolorspace").equals("off")) {
                try {
                    csMap = new ColorSpace(in, hd, pl);
                    channels = hd.
                            createChannelDefinitionMapper(ictransf, csMap);
                    resampled = hd.createResampler(channels, csMap);
                    palettized = hd.
                            createPalettizedColorSpaceMapper(resampled, csMap);
                    color = hd.createColorSpaceMapper(palettized, csMap);
                } catch (IllegalArgumentException e) {
                    error("Could not instantiate ICC profiler" +
                            ((e.getMessage() != null) ?
                            (":\n" + e.getMessage()) : ""), 1);
                    return;
                } catch (ColorSpaceException e) {
                    error("error processing jp2 colorspace information" +
                            ((e.getMessage() != null) ?
                            (": " + e.getMessage()) : "    "), 1);
                    return;
                }
            } else { // Skip colorspace mapping
                color = ictransf;
            }

            // This is the last image in the decoding chain and should be
            // assigned by the last transformation:
            decodedImage = color;
            if (color == null) {
                decodedImage = ictransf;
            }

            // **** Report info ****
            reconstructedResolutionLevel = res;
            targetResolutionLevel = decSpec.dls.getMin();
            reconstructedImageWidth = breader.getImgWidth(res);
            reconstructedImageHeight = breader.getImgHeight(res);
            targetBitrate = breader.getTargetRate();

            // **** Decode and write/display result ****
            // Now create the image and decode. Use a low priority for
            // this so as not to block other threads.
            Thread.currentThread().setPriority(Thread.MIN_PRIORITY + 1);
            img = BlkImgDataSrcImageProducer.createImage(decodedImage, isp);
            isp.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
            isp.setImage(img);
            isp.setCursor(Cursor.
                    getPredefinedCursor(Cursor.DEFAULT_CURSOR));

            // Check the image status, every 100ms, until it is finished.
            if (win != null) {
                int status;
                do {
                    status = isp.checkImage(img, null);
                    if ((status & ImageObserver.ERROR) != 0) {
                        error("An unknown error occurred while " +
                                "producing the image", 2);
                        return;
                    } else if ((status & ImageObserver.ABORT) != 0) {
                        error("Image production was aborted for " +
                                "some unknown reason", 2);
                    } else if ((status & ImageObserver.ALLBITS) != 0) {
                        ImgMouseListener iml = new ImgMouseListener(isp);
                        isp.addMouseListener(iml);
                        isp.addMouseMotionListener(iml);
                    } else { // Check again in 100 ms
                        try {
                            Thread.sleep(100);
                        } catch (InterruptedException e) {
                        }
                    }
                } while ((status &
                        (ImageObserver.ALLBITS | ImageObserver.ABORT |
                        ImageObserver.ERROR)) == 0);
            }

        } catch (IllegalArgumentException e) {
            error(e.getMessage(), 2);
            return;
        } catch (Error e) {
            if (e.getMessage() != null) {
                error(e.getMessage(), 2);
            } else {
                error("An error has occured during decoding.", 2);
            }
            return;
        } catch (RuntimeException e) {
            if (e.getMessage() != null) {
                error("An uncaught runtime exception has occurred:\n" +
                        e.getMessage(), 2);
            } else {
                error("An uncaught runtime exception has occurred.", 2);
            }
            return;
        } catch (Throwable e) {
            error("An uncaught exception has occurred.", 2);
            return;
        }
    }

    /**
     * Returns all the parameters used in the decoding chain. It calls
     * parameter from each module and store them in one array (one row per
     * parameter and 4 columns).
     *
     * @return All decoding parameters
     *
     * @see #getParameterInfo
     * */
    private static String[][] getAllParameters() {
        Vector vec = new Vector();
        int i;

        String[][] str = BitstreamReaderAgent.getParameterInfo();
        if (str != null) for (i = str.length - 1; i >= 0; i--) vec.addElement(str[i]);

        str = EntropyDecoder.getParameterInfo();
        if (str != null) for (i = str.length - 1; i >= 0; i--) vec.addElement(str[i]);

        str = ROIDeScaler.getParameterInfo();
        if (str != null) for (i = str.length - 1; i >= 0; i--) vec.addElement(str[i]);

        str = Dequantizer.getParameterInfo();
        if (str != null) for (i = str.length - 1; i >= 0; i--) vec.addElement(str[i]);

        str = InvCompTransf.getParameterInfo();
        if (str != null) for (i = str.length - 1; i >= 0; i--) vec.addElement(str[i]);

        str = HeaderDecoder.getParameterInfo();
        if (str != null) for (i = str.length - 1; i >= 0; i--) vec.addElement(str[i]);

        str = ICCProfiler.getParameterInfo();
        if (str != null) for (i = str.length - 1; i >= 0; i--) vec.addElement(str[i]);

        str = getParameterInfo();
        if (str != null) for (i = str.length - 1; i >= 0; i--) vec.addElement(str[i]);

        str = new String[vec.size()][4];
        if (str != null)
            for (i = str.length - 1; i >= 0; i--)
                str[i] = (String[]) vec.elementAt(i);

        return str;
    }

    /**
     * Prints the error message 'msg' to errorMessage variable
     * and sets the exitCode to 'code'. An exit code different than 0
     * indicates that there where problems.
     *
     * @param msg The error message
     *
     * @param code The exit code to set
     * */
    private void error(String msg, int code) {
        exitCode = code;
        errorMessage = msg;
    }

    /**
     * Return the information found in the COM marker segments encountered in
     * the decoded codestream.
     * */
    public String[] getCOMInfo() {
        if (hi == null) { // The codestream has not been read yet
            return null;
        }

        int nCOMMarkers = hi.getNumCOM();
        Enumeration com = hi.com.elements();
        String[] infoCOM = new String[nCOMMarkers];
        for (int i = 0; i < nCOMMarkers; i++) {
            infoCOM[i] = com.nextElement().toString();
        }
        return infoCOM;
    }

}
