package habibviewer.jpeg2000;

/*
 * CVS identifier:
 *
 * $Id: habibviewer.jpeg2000.Encoder.java,v 1.68 2002/05/22 14:59:08 grosbois Exp $
 *
 * Class:                   habibviewer.jpeg2000.Encoder
 *
 * Description:             The encoder object
 *
 *                          [from CmdLnEncoder, Diego SANTA CRUZ, May-19-1999]
 *
 *
 * COPYRIGHT:
 *
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 *
 * Copyright (c) 1999/2000 JJ2000 Partners.
 * */

import jj2000.j2k.quantization.quantizer.*;
import jj2000.j2k.image.forwcomptransf.*;
import jj2000.j2k.codestream.writer.*;
import jj2000.j2k.fileformat.writer.*;
import jj2000.j2k.wavelet.analysis.*;
import jj2000.j2k.entropy.encoder.*;
import jj2000.j2k.image.input.*;
import jj2000.j2k.roi.encoder.*;
import jj2000.j2k.image.*;
import jj2000.j2k.util.*;
import jj2000.j2k.*;
import jj2000.j2k.encoder.EncoderSpecs;
import jj2000.disp.ImgScrollPane;

import java.util.*;
import java.io.*;

/**
 * This class is the main class of JJ2000's encoder. It instantiates all
 * objects of the chain and launchs the encoding process. It then writes the
 * header and the compressed bit stream to the output file. Finally,
 * packed packet headers (through codestream post-manipulation) and
 * file-format may be created if needed.
 *
 * <p>First the encoder should be initialized with a ParameterList object
 * provided through the constructor. Then, the run() method is invoked and the
 * encoder executes. The exit code of the class can be obtained with the
 * getExitCode() method, after the constructor and after the run method. A
 * non-zero value indicates that an error has occurred.</p>
 *
 * <p>The modules are inserted in the encoding chain with the following
 * order:</p>
 *
 * <ul>
 * <li>ImgReader</li>
 * <li>ImgDataJoiner (if multiple image readers)</li>
 * <li>Tiler</li>
 * <li>ForwCompTransf</li>
 * <li>ImgDataConverter</li>
 * <li>ForwardWT</li>
 * <li>Quantizer</li>
 * <li>ROIScaler</li>
 * <li>EntropyCoder</li>
 * <li>PostCompRateAllocator</li>
 * </ul>
 *
 * <p>The encoder uses a pull model. This means that the last module
 * (PostCompRateAllocator) requests data from its source (EntropyCoder),
 * ...</p>
 *
 * <p>Writing of the codestream writing (header+bit stream) is realized by
 * HeaderEncoder and CodestreamWriter modules.</p>
 *
 * <p>Packed packet headers and file-format creation are carried out by
 * CodestreamManipulator and FileFormatWriter modules respectively.</p>
 *
 * <p>Many modules of the encoder may behave differently depending on the
 * tile-component. The specifications of their behaviour are kept in
 * specialized modules extending ModuleSpec class. All these modules are
 * accessible through an instance of EncoderSpecs class.</p>
 *
 * @see ImgReader
 * @see ImgDataJoiner
 * @see ForwCompTransf
 * @see Tiler
 * @see ImgDataConverter
 * @see ForwardWT
 * @see Quantizer
 * @see ROIScaler
 * @see EntropyCoder
 * @see PostCompRateAllocator
 * @see HeaderEncoder
 * @see CodestreamWriter
 * @see CodestreamManipulator
 * @see FileFormatWriter
 * @see ModuleSpec
 * @see jj2000.j2k.encoder.EncoderSpecs
 * */
public class Encoder implements Runnable {

    /** The exit code of the run method */
    public int exitCode;

    public String errorMessage = "";

    /** The parameter list (arguments) */
    private ParameterList pl;

    /** The default parameter list (arguments) */
    private ParameterList defpl;

    ImgScrollPane isp;
    String outname;
    BlkImgDataSrc imgsrc;


    /** The valid list of options prefixes */
    public final static char vprfxs[] = {
        ForwCompTransf.OPT_PREFIX, // Mixer module
        AnWTFilter.OPT_PREFIX, // Filters type spec
        ForwardWT.OPT_PREFIX, // Wavelets module
        Quantizer.OPT_PREFIX, // Quantizer module
        ROIScaler.OPT_PREFIX, // ROI module
        HeaderEncoder.OPT_PREFIX, // HeaderEncoder module
        EntropyCoder.OPT_PREFIX, // Coding modules
        PostCompRateAllocator.OPT_PREFIX, // Rate allocator
        PktEncoder.OPT_PREFIX, // Packet encoder
    };

    /** The parameter information for this class */
    private final static String[][] pinfo = {
        {"debug", null,
         "Print debugging messages when an error is encountered.", "off"},
        {"disable_jp2_extension", "[on|off]",
         "JJ2000 automatically adds .jp2 extension when using 'file_format'" +
            "option. This option disables it when on.", "off"},
        {"file_format", "[on|off]",
         "Puts the JPEG 2000 codestream in a JP2 file format wrapper.", "off"},
        {"pph_tile", "[on|off]",
         "Packs the packet headers in the tile headers.", "off"},
        {"pph_main", "[on|off]",
         "Packs the packet headers in the main header.", "off"},
        {"pfile", "<filename of arguments file>",
         "Loads the arguments from the specified file. Arguments that are " +
            "specified on the command line override the ones from the file.\n" +
            "The arguments file is a simple text file with one argument per " +
            "line of the following form:\n" +
            "  <argument name>=<argument value>\n" +
            "If the argument is of boolean type (i.e. its presence turns a " +
            "feature on), then the 'on' value turns it on, while the 'off' " +
            "value turns it off. The argument name does not include the '-' " +
            "or '+' character. Long lines can be broken into several lines " +
            "by terminating them with '\'. Lines starting with '#' are " +
            "considered as comments. This option is not recursive: any 'pfile' " +
            "argument appearing in the file is ignored.", null},
        {"tile_parts", "<packets per tile-part>",
         "This option specifies the maximum number of packets to have in " +
            "one tile-part. 0 means include all packets in first tile-part " +
            "of each tile", "0"},
        {"tiles", "<nominal tile width> <nominal tile height>",
         "This option specifies the maximum tile dimensions to use. " +
            "If both dimensions are 0 then no tiling is used.", "0 0"},
        {"ref", "<x> <y>",
         "Sets the origin of the image in the canvas system. It sets the " +
            "coordinate of the top-left corner of the image reference grid, " +
            "with respect to the canvas origin", "0 0"},
        {"tref", "<x> <y>",
         "Sets the origin of the tile partitioning on the reference grid, " +
            "with respect to the canvas origin. The value of 'x' ('y') " +
            "specified can not be larger than the 'x' one specified in the ref " +
            "option.", "0 0"},
        {"rate", "<output bitrate in bpp>",
         "This is the output bitrate of the codestream in bits per pixel." +
            " When equal to -1, no image information (beside quantization " +
            "effects) is discarded during compression.\n" +
            "Note: In the case where '-file_format' option is used, the " +
            "resulting file may have a larger bitrate.", "-1"},
        {"lossless", "[on|off]",
         "Specifies a lossless compression for the encoder. This options" +
            " is equivalent to use reversible quantization ('-Qtype " +
            "reversible')" +
            " and 5x3 wavelet filters pair ('-Ffilters w5x3'). Note that " +
            "this option cannot be used with '-rate'. When this option is " +
            "off, the quantization type and the filters pair is defined by " +
            "'-Qtype' and '-Ffilters' respectively.", "off"},
        {"i", "<image file> [,<image file> [,<image file> ... ]]",
         "Mandatory argument. This option specifies the name of the input " +
            "image files. If several image files are provided, they have to be" +
            " separated by commas in the command line. Supported formats are " +
            "PGM (raw), PPM (raw) and PGX, " +
            "which is a simple extension of the PGM file format for single " +
            "component data supporting arbitrary bitdepths. If the extension " +
            "is '.pgm', PGM-raw file format is assumed, if the extension is " +
            "'.ppm', PPM-raw file format is assumed, otherwise PGX file " +
            "format is assumed. PGM and PPM files are assumed to be 8 bits " +
            "deep. A multi-component image can be specified by either " +
            "specifying several PPM and/or PGX files, or by specifying one " +
            "PPM file.", null},
        {"o", "<file name>",
         "Mandatory argument. This option specifies the name of the output " +
            "file to which the codestream will be written.", null},
        {"verbose", null,
         "Prints information about the obtained bit stream.", "on"},
        {"v", "[on|off]",
         "Prints version and copyright information.", "off"},
        {"u", "[on|off]",
         "Prints usage information. " +
            "If specified all other arguments (except 'v') are ignored", "off"},
    };

    /**
     * Instantiates an encoder object, width the ParameterList object given as
     * argument. It also retrieves the default ParameterList.
     *
     * */
    public Encoder(BlkImgDataSrc imgSrc, String outputFile, float bitrate, int layers) {
        this.imgsrc = imgSrc;
        this.outname = outputFile;

        //ParameterList defaultParameterList;
        defpl = new ParameterList();
        String[][] param = getAllParameters();
        for (int i = param.length - 1; i >= 0; i--) {
            if (param[i][3] != null)
                defpl.put(param[i][0], param[i][3]);
        }

        pl = new ParameterList(defpl);
        pl.put("o", outputFile);
        if (bitrate != 0) {
            pl.put("rate", Float.toString(bitrate));
        } else {
            pl.put("lossless", "on");
        }

        if (layers > 1) {
            String args = "";
            if (bitrate > 0) {
                for (int i = 1; i <= layers; i++)
                    args += i*bitrate/layers+ " ";
            } else
                args += 0.1 + " +" + (layers-2);
            pl.put("Alayers", args);
        } else {
            pl.put("Alayers", "sl");
        }
    }

    /**
     * Returns the exit code of the class. This is only initialized after the
     * constructor and when the run method returns.
     *
     * @return The exit code of the constructor and the run() method.
     * */
    public int getExitCode() {
        return exitCode;
    }

    /**
     * Runs the encoder. After completion the exit code is set, a non-zero
     * value indicates that an error ocurred.
     *
     * @see #getExitCode
     * */
    public void run() {
        boolean useFileFormat = false;
        boolean pphTile = false;
        boolean pphMain = false;
        boolean tempSop = false;
        boolean tempEph = false;
        StreamTokenizer stok;
        StringTokenizer sgtok;
        int ncomp;
        boolean imsigned[];
        int tw,th;
        int refx,refy;
        int trefx,trefy;
        int pktspertp;
        Tiler imgtiler;
        ForwCompTransf fctransf;
        ImgDataConverter converter;
        EncoderSpecs encSpec;
        ForwardWT dwt;
        Quantizer quant;
        ROIScaler rois;
        EntropyCoder ecoder;
        PostCompRateAllocator ralloc;
        HeaderEncoder headenc;
        CodestreamWriter bwriter;
        FileFormatWriter ffw;
        float rate;
        int fileLength;

        try {
            // Check that we have the mandatory parameters
            if (outname == null) {
                error("Mandatory output file is missing", 2);
                return;
            }

            if (pl.getParameter("file_format").equals("on")) {
                useFileFormat = true;
            }

            if (useFileFormat) {
                String outext = null;
                String outns = outname;
                if (outname.lastIndexOf('.') != -1) {
                    outext = outname.substring(outname.lastIndexOf('.'),
                            outname.length());
                    outns = outname.substring(0, outname.lastIndexOf('.'));
                }

                if (outext == null || !outext.equalsIgnoreCase(".jp2")) {
                    if (!pl.getBooleanParameter("disable_jp2_extension")) {
                        outname = outns + ".jp2";
                    }
                }
            }

            if (pl.getParameter("tiles") == null) {
                error("No tiles option specified", 2);
                return;
            }

            if (pl.getParameter("pph_tile").equals("on")) {
                pphTile = true;

                if (pl.getParameter("Psop").equals("off")) {
                    pl.put("Psop", "on");
                    tempSop = true;
                }
                if (pl.getParameter("Peph").equals("off")) {
                    pl.put("Peph", "on");
                    tempEph = true;
                }
            }

            if (pl.getParameter("pph_main").equals("on")) {
                pphMain = true;

                if (pl.getParameter("Psop").equals("off")) {
                    pl.put("Psop", "on");
                    tempSop = true;
                }
                if (pl.getParameter("Peph").equals("off")) {
                    pl.put("Peph", "on");
                    tempEph = true;
                }
            }

            if (pphTile && pphMain)
                error("Can't have packed packet headers in both main and" +
                        " tile headers", 2);

            if (pl.getBooleanParameter("lossless") &&
                    pl.getParameter("rate") != null &&
                    pl.getFloatParameter("rate") != defpl.getFloatParameter("rate"))
                throw new IllegalArgumentException("Cannot use '-rate' and " +
                        "'-lossless' option at " +
                        " the same time.");

            if (pl.getParameter("rate") == null) {
                error("Target bitrate not specified", 2);
                return;
            }
            try {
                rate = pl.getFloatParameter("rate");
                if (rate == -1) {
                    rate = Float.MAX_VALUE;
                }
            } catch (NumberFormatException e) {
                error("Invalid value in 'rate' option: " +
                        pl.getParameter("rate"), 2);
                return;
            }
            try {
                pktspertp = pl.getIntParameter("tile_parts");
                if (pktspertp != 0) {
                    if (pl.getParameter("Psop").equals("off")) {
                        pl.put("Psop", "on");
                        tempSop = true;
                    }
                    if (pl.getParameter("Peph").equals("off")) {
                        pl.put("Peph", "on");
                        tempEph = true;
                    }
                }
            } catch (NumberFormatException e) {
                error("Invalid value in 'tile_parts' option: " +
                        pl.getParameter("tile_parts"), 2);
                return;
            }

            // **** ImgReader ****
            ncomp = 3;

            // **** Tiler ****
            // get nominal tile dimensions
            stok =
                    new StreamTokenizer(new StringReader(pl.
                    getParameter("tiles")));
            stok.eolIsSignificant(false);

            stok.nextToken();
            if (stok.ttype != StreamTokenizer.TT_NUMBER) {
                error("An error occurred while parsing the tiles option: " +
                        pl.getParameter("tiles"), 2);
                return;
            }
            tw = (int) stok.nval;
            stok.nextToken();
            if (stok.ttype != StreamTokenizer.TT_NUMBER) {
                error("An error occurred while parsing the tiles option: " +
                        pl.getParameter("tiles"), 2);
                return;
            }
            th = (int) stok.nval;

            // Get image reference point
            sgtok = new StringTokenizer(pl.getParameter("ref"));
            try {
                refx = Integer.parseInt(sgtok.nextToken());
                refy = Integer.parseInt(sgtok.nextToken());
            } catch (NoSuchElementException e) {
                throw
                        new IllegalArgumentException("Error while parsing 'ref' " +
                        "option");
            } catch (NumberFormatException e) {
                throw new IllegalArgumentException("Invalid number type in " +
                        "'ref' option");
            }
            if (refx < 0 || refy < 0) {
                throw new IllegalArgumentException("Invalid value in 'ref' " +
                        "option ");
            }

            // Get tiling reference point
            sgtok = new StringTokenizer(pl.getParameter("tref"));
            try {
                trefx = Integer.parseInt(sgtok.nextToken());
                trefy = Integer.parseInt(sgtok.nextToken());
            } catch (NoSuchElementException e) {
                throw
                        new IllegalArgumentException("Error while parsing 'tref' " +
                        "option");
            } catch (NumberFormatException e) {
                throw new IllegalArgumentException("Invalid number type in " +
                        "'tref' option");
            }
            if (trefx < 0 || trefy < 0 || trefx > refx || trefy > refy) {
                throw new IllegalArgumentException("Invalid value in 'tref' " +
                        "option ");
            }

            // Instantiate tiler
            try {
                imgtiler = new Tiler(imgsrc, refx, refy, trefx, trefy, tw, th);
            } catch (IllegalArgumentException e) {
                error("Could not tile image" +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }
            int ntiles = imgtiler.getNumTiles();

            // **** Encoder specifications ****
            encSpec = new EncoderSpecs(ntiles, ncomp, imgsrc, pl);

            // **** Component transformation ****
            try {
                fctransf = new ForwCompTransf(imgtiler, encSpec);
            } catch (IllegalArgumentException e) {
                error("Could not instantiate forward component " +
                        "transformation" +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }

            // **** ImgDataConverter ****
            converter = new ImgDataConverter(fctransf);


            // **** ForwardWT ****
            try {
                dwt = ForwardWT.createInstance(converter, pl, encSpec);
            } catch (IllegalArgumentException e) {
                error("Could not instantiate wavelet transform" +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }

            // **** Quantizer ****
            try {
                quant = Quantizer.createInstance(dwt, encSpec);
            } catch (IllegalArgumentException e) {
                error("Could not instantiate quantizer" +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }

            // **** ROIScaler ****
            try {
                rois = ROIScaler.createInstance(quant, pl, encSpec);
            } catch (IllegalArgumentException e) {
                error("Could not instantiate ROI scaler" +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }

            // **** EntropyCoder ****
            try {
                ecoder = EntropyCoder.createInstance(rois, pl, encSpec.cblks,
                        encSpec.pss, encSpec.bms,
                        encSpec.mqrs, encSpec.rts,
                        encSpec.css, encSpec.sss,
                        encSpec.lcs, encSpec.tts);
            } catch (IllegalArgumentException e) {
                error("Could not instantiate entropy coder" +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }

            // **** CodestreamWriter ****
            try {
                // Rely on rate allocator to limit amount of data
                bwriter = new FileCodestreamWriter(outname, Integer.MAX_VALUE);
            } catch (IOException e) {
                error("Could not open output file" +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }

            // **** Rate allocator ****
            try {
                ralloc = PostCompRateAllocator.createInstance(ecoder, pl, rate,
                        bwriter, encSpec);
            } catch (IllegalArgumentException e) {
                error("Could not instantiate rate allocator" +
                        ((e.getMessage() != null) ?
                        (":\n" + e.getMessage()) : ""), 2);
                return;
            }
            //////
            imsigned = new boolean[ncomp];
            imsigned[0] = true;

            // **** HeaderEncoder ****
            headenc = new HeaderEncoder(imgsrc, imsigned, dwt, imgtiler, encSpec,
                    rois, ralloc, pl);
            ralloc.setHeaderEncoder(headenc);

            // **** Write header to be able to estimate header overhead ****
            headenc.encodeMainHeader();

            // **** Initialize rate allocator, with proper header
            // overhead. This will also encode all the data ****
            ralloc.initialize();

            // **** Write header (final) ****
            headenc.reset();
            headenc.encodeMainHeader();

            // Insert header into the codestream
            bwriter.commitBitstreamHeader(headenc);

            // **** Now do the rate-allocation and write result ****
            ralloc.runAndWrite();

            // **** Done ****
            bwriter.close();

            // **** Calculate file length ****
            fileLength = bwriter.getLength();

            // **** Tile-parts and packed packet headers ****
            if (pktspertp > 0 || pphTile || pphMain) {
                try {
                    CodestreamManipulator cm = new
                            CodestreamManipulator(outname, ntiles, pktspertp,
                                    pphMain, pphTile, tempSop, tempEph);
                    fileLength += cm.doCodestreamManipulation();
                } catch (IOException e) {
                    error("Error while creating tileparts or packed packet" +
                            " headers" +
                            ((e.getMessage() != null) ?
                            (":\n" + e.getMessage()) : ""), 2);
                    if (pl.getParameter("debug").equals("on")) {
                        e.printStackTrace();
                    } else {
                        error("Use '-debug' option for more details", 2);
                    }
                    return;
                }
            }

            // **** File Format ****
            if (useFileFormat) {
                try {
                    int nc = imgsrc.getNumComps();
                    int[] bpc = new int[nc];
                    for (int comp = 0; comp < nc; comp++) {
                        bpc[comp] = imgsrc.getNomRangeBits(comp);
                    }

                    ffw = new FileFormatWriter(outname, imgsrc.getImgHeight(),
                            imgsrc.getImgWidth(), nc, bpc,
                            fileLength);
                    fileLength += ffw.writeFileFormat();
                } catch (IOException e) {
                    throw new Error("Error while writing JP2 file format");
                }
            }

        } catch (IllegalArgumentException e) {
            error(e.getMessage(), 2);
        } catch (Error e) {
            error("An uncaught error has occurred: " + e.getMessage(), 2);
        } catch (RuntimeException e) {
            error("An uncaught runtime exception has occurred: " +
                    e.getMessage(), 2);
        } catch (Throwable e) {
            error("An unchecked exception has occurred: " +
                    e.getMessage(), 2);
            return;
        }
    }

    /**
     * Returns the parameters that are used in this class and
     * implementing classes. It returns a 2D String array. Each of the
     * 1D arrays is for a different option, and they have 4
     * elements. The first element is the option name, the second one
     * is the synopsis, the third one is a long description of what
     * the parameter is and the fourth is its default value. The
     * synopsis or description may be 'null', in which case it is
     * assumed that there is no synopsis or description of the option,
     * respectively. Null may be returned if no options are supported.
     *
     * @return the options name, their synopsis and their explanation,
     * or null if no options are supported.
     * */
    private static String[][] getParameterInfo() {
        return pinfo;
    }

    /**
     * Returns all the parameters used in the encoding chain. It calls
     * parameter from each module and store them in one array (one row
     * per parameter and 4 columns).
     *
     * @return All encoding parameters
     *
     * @see #getParameterInfo
     * */
    private static String[][] getAllParameters() {
        Vector vec = new Vector();

        String[][] str = getParameterInfo();
        if (str != null)
            for (int i = str.length - 1; i >= 0; i--)
                vec.addElement(str[i]);

        str = ForwCompTransf.getParameterInfo();
        if (str != null)
            for (int i = str.length - 1; i >= 0; i--)
                vec.addElement(str[i]);

        str = AnWTFilter.getParameterInfo();
        if (str != null)
            for (int i = str.length - 1; i >= 0; i--)
                vec.addElement(str[i]);

        str = ForwardWT.getParameterInfo();
        if (str != null)
            for (int i = str.length - 1; i >= 0; i--)
                vec.addElement(str[i]);

        str = Quantizer.getParameterInfo();
        if (str != null)
            for (int i = str.length - 1; i >= 0; i--)
                vec.addElement(str[i]);

        str = ROIScaler.getParameterInfo();
        if (str != null)
            for (int i = str.length - 1; i >= 0; i--)
                vec.addElement(str[i]);

        str = EntropyCoder.getParameterInfo();
        if (str != null)
            for (int i = str.length - 1; i >= 0; i--)
                vec.addElement(str[i]);

        str = HeaderEncoder.getParameterInfo();
        if (str != null)
            for (int i = str.length - 1; i >= 0; i--)
                vec.addElement(str[i]);

        str = PostCompRateAllocator.getParameterInfo();
        if (str != null)
            for (int i = str.length - 1; i >= 0; i--)
                vec.addElement(str[i]);

        str = PktEncoder.getParameterInfo();
        if (str != null)
            for (int i = str.length - 1; i >= 0; i--)
                vec.addElement(str[i]);

        str = new String[vec.size()][4];
        if (str != null)
            for (int i = str.length - 1; i >= 0; i--)
                str[i] = (String[]) vec.elementAt(i);

        return str;
    }

    /**
     * Prints the error message 'msg' to errorMessage variable
     * and sets the exitCode to 'code'. An exit code different than 0
     * indicates that there where problems.
     *
     * @param msg The error message
     *
     * @param code The exit code to set
     * */
    private void error(String msg, int code) {
        exitCode = code;
        errorMessage = msg;
    }
}
